'''
An encryption plugin for Elixir utilizing the excellent PyCrypto library, which
can be downloaded here: http://www.amk.ca/python/code/crypto

Values for columns that are specified to be encrypted will be transparently
encrypted and safely encoded for storage in a unicode column using the powerful
and secure Blowfish Cipher using a specified "secret" which can be passed into
the plugin at class declaration time.

Example usage:

.. sourcecode:: python

    from elixir import *
    from elixir.ext.encrypted import acts_as_encrypted
    
    class Person(Entity):
        name = Field(Unicode)
        password = Field(Unicode)
        ssn = Field(Unicode)
        acts_as_encrypted(for_columns=['password', 'ssn'], 
                          with_secret='secret')

The above Person entity will automatically encrypt and decrypt the password and
ssn columns on save, update, and load.  Different secrets can be specified on
an entity by entity basis, for added security.
'''

import sys
import os
from Crypto.Cipher          import Blowfish, AES
from elixir.statements      import Statement
from sqlalchemy.orm         import MapperExtension, EXT_PASS

__all__ = ['acts_as_encrypted']
__doc_all__ = []


#
# encryption and decryption functions
#
# WARNING!!! Blowfish encryption method is vulnerable to attacks
# because it doesn't properly use random seed. It is provided just for
# backward compatibility needed to migrate data. Use AES instead!
def encrypt_value(value, secret):
    return Blowfish.new(secret, Blowfish.MODE_CFB) \
                   .encrypt(value).encode('string_escape')

def decrypt_value(value, secret):
    return Blowfish.new(secret, Blowfish.MODE_CFB) \
                   .decrypt(value.decode('string_escape'))

# Crypto.Cipher.AES is AES128
def encrypt_value_aes(value, secret):
    iv = os.urandom(AES.block_size)

    pad_len = AES.block_size - len(value) % AES.block_size
    padded_value = value + pad_len * chr(pad_len)
    res = iv + AES.new(secret, AES.MODE_CBC, iv).encrypt(padded_value)
    return res.encode('string_escape')

def decrypt_value_aes(value, secret):
    value = value.decode('string_escape')
    iv = value[:AES.block_size]
    encrypted = value[AES.block_size:]

    padded_value = AES.new(secret, AES.MODE_CBC, iv).decrypt(encrypted)
    pad_len = ord(padded_value[-1])
    assert pad_len >= 1 and pad_len <= AES.block_size
    return padded_value[:-pad_len]

#
# acts_as_encrypted statement
#

class ActsAsEncrypted(object):    

    def __init__(self, entity, for_fields=[], with_secret='abcdef', with_aes=False):
        if not with_aes:
            sys.stderr.write("""******* WARNING!!! ********
Blowfish encryption method is vulnerable to attacks.
Migrate your data and use with_aes=True\n""")

        def perform_encryption(instance, decrypt=False):
            for column_name in for_fields:
                current_value = getattr(instance, column_name)
                if current_value:
                    if decrypt:
                        if with_aes:
                            new_value = decrypt_value_aes(current_value, with_secret)
                        else:
                            new_value = decrypt_value(current_value, with_secret)
                    else:
                        if with_aes:
                            new_value = encrypt_value_aes(current_value, with_secret)
                        else:
                            new_value = encrypt_value(current_value, with_secret)
                    setattr(instance, column_name, new_value)
        
        def perform_decryption(instance):
            perform_encryption(instance, decrypt=True)
        
        class EncryptedMapperExtension(MapperExtension):            
            
            def before_insert(self, mapper, connection, instance):
                perform_encryption(instance)
                return EXT_PASS
            
            def before_update(self, mapper, connection, instance):       
                perform_encryption(instance)
                return EXT_PASS
            
            def populate_instance(self, mapper, selectcontext, row, instance, 
                                  *args, **kwargs):
                mapper.populate_instance(selectcontext, instance, row, 
                                         *args, **kwargs)
                perform_decryption(instance)
                return True
        
        # make sure that the entity's mapper has our mapper extension
        entity._descriptor.add_mapper_extension(EncryptedMapperExtension())


acts_as_encrypted = Statement(ActsAsEncrypted)

